using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using System.Query;

namespace Iterators
{
    public delegate bool Sorter<T>(T rhs, T lhs);

    public static class Sequence
    {
        #region Collect
        /// <summary>
        /// Collects elements from a source array into a new array following a transformation
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="block"></param>
        /// <returns></returns>
        public static IEnumerable<T> Collect<T>(this IEnumerable<T> source, Func<T, T> collect)
        {
            foreach (T item in source)
            {
                yield return collect(item);
            }
        }
        #endregion

        #region Each
        /// <summary>
        /// Each performs an operation on each item in a collection. Note that unless the item
        /// is passed by reference this will not modify the item in the collection itself
        /// ForEach(Action<T> action) already provides similar functionality
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="action"></param>
        public static void Each<T>(this IEnumerable<T> source, Action<T> action)
        {
            foreach (T item in source)
            {
                action(item);
            }
        }
        #endregion

        #region Find
        /// <summary>
        /// This method returns a list of all elements that match a given query
        /// Note that FindAll already does this for some generics
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="query"></param>
        /// <returns></returns>
        public static IEnumerable<T> Find<T>(this IEnumerable<T> source, Predicate<T> query)
        {
            foreach (T item in source)
            {
                if (query(item))
                    yield return item;
            }
        }
        #endregion

        #region Sort
        /// <summary>
        /// The sort algorithm here is not particulary sophisticated
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="source"></param>
        /// <param name="comparer"></param>
        /// <returns></returns>
        public static IEnumerable<T> Sort<T>(this IEnumerable<T> source, Sorter<T> comparison)
        {
            List<T> sorted = new List<T>();

            foreach (T item in source)
            {
                if (sorted.Count == 0)
                    sorted.Add(item);
                else
                {
                    for (int i = 0; i < sorted.Count; i++)
                    {
                        if (comparison(sorted[i], item))
                        {
                            sorted.Insert(i, item);
                            break;
                        }
                    }
                }
            }
            return sorted;
        }
        #endregion

        #region Down To
        /// <summary>
        /// Down to just loops over a range and carries out an action
        /// </summary>
        /// <param name="upperBound"></param>
        /// <param name="lowerBound"></param>
        /// <param name="action"></param>
        public static void DownTo(this int upperBound, int lowerBound, Action<int> action)
        {
            for (int i = upperBound; i >= lowerBound; i--)
            {
                action(i);
            }
        }

        /// <summary>
        /// Helper method for printing any object
        /// </summary>
        /// <param name="obj"></param>
        public static void Print(this object obj)
        {
            Console.WriteLine(obj);
        }
        #endregion

        #region Step
        /// <summary>
        /// Again this is like downto, but we just count up
        /// </summary>
        /// <param name="from"></param>
        /// <param name="to"></param>
        /// <param name="action"></param>
        public static void Step(this int from, int to, Action<int> action)
        {
            for (int i = from; i <= to; i++)
            {
                action(i);
            }
        }
        #endregion

        #region Times
        /// <summary>
        /// Do action a number of times
        /// </summary>
        /// <param name="times"></param>
        /// <param name="action"></param>
        public static void Times(this int times, Action<int> action)
        {
            for (int i = 1; i <= times; i++)
            {
                action(i);
            }
        }
        #endregion

        #region Up To
        /// <summary>
        /// From beginning down to, do this action
        /// </summary>
        /// <param name="upperBound"></param>
        /// <param name="lowerBound"></param>
        /// <param name="action"></param>
        public static void UpTo(this int lowerBound, int upperBound, Action<int> action)
        {
            for (int i = lowerBound; i <= upperBound; i++)
            {
                action(i);
            }
        }

        #endregion

    }
}
