using System;
using System.Collections.Generic;
using System.Text;
using System.Query;

//A lot of this code is derived from Don Box at: http://www.pluralsight.com/blogs/dbox/archive/2005/04/24/7690.aspx
//Don also has exlusive and inclusive ranges, I have omitted those for simplicity
//We have converted from Converter<T,T> to C#3.0 Func<T,T> for clarity over the successor function, which is not doing conversions


namespace Extensions
{
    public static class Range
    {
        public static Predicate<T> IsIn<T>(T start, T finish, Comparison<T> comp)
        {
            return (T item) => comp(start, item) <= 0 && comp(item, finish) <= 0;
        }

        public static Predicate<T> IsIn<T>(T start, T finish) where T : IComparable<T>
        {
            return IsIn<T>(start, finish, comp);
        }

        #region Range Generators
        public static IEnumerable<T> Create<T>(T start, T finish, Func<T, T> succ, Comparison<T> comp)
        {
            T value = start;
            int finished = comp(value, finish);
            while (finished <= 0)
            {
                yield return value;
                if (finished == 0)
                    break;
                else
                    value = succ(value);

                finished = comp(value, finish);
            }
        }

        internal static int comp<T>(T a, T b) where T : IComparable<T>
        {
            if (a != null)
                return a.CompareTo(b);
            return (b == null) ? 0 : -1;
        }

        public static IEnumerable<T> Create<T>(T start, T finish, Func<T, T> succ) where T : IComparable<T>
        {
            return Create<T>(start, finish, succ, comp);
        }
        #endregion

        #region Successor Functions
        /// We need some way to iterate through a range when we generate i.e. figure out what the next element
        /// in the sequence should be. So for any type we want to create ranges of, we need to defind
        /// a successor method. Note that we can only do this for sequences where we either have all the elements
        /// or where we can predict the sequence.

        #region Integer types successor generators
        internal static long succ(long val) { return val + 1; }
        internal static int succ(int val) { return val + 1; }
        internal static short succ(short val) { return (short)(val + 1); }
        internal static sbyte succ(sbyte val) { return (sbyte)(val + 1); }
         internal static ulong succ(ulong val) { return val + 1; }
        internal static uint succ(uint val) { return val + 1; }
        internal static ushort succ(ushort val) { return (ushort)(val + 1); }
        internal static byte succ(byte val) { return (byte)(val + 1); }
 
        #endregion

        #region Character types successor generators
        internal static char succ(char val) { return (char)(val + 1); }
        #endregion

        #region DateTime successor generator
        internal static DateTime succ(DateTime val) { return val.AddDays(1);}
        #endregion

        #region string successor generator
        internal static string succ(string val)
        {
            int lastAlphaNumeric = -1;
            for (int i = val.Length - 1; i >= 0 && lastAlphaNumeric == -1; i--)
            {
                if (char.IsLetterOrDigit(val[i]))
                    lastAlphaNumeric = i;
            }
            if (lastAlphaNumeric == val.Length - 1 || lastAlphaNumeric == -1)
                return succ(val, val.Length);
            return succ(val, lastAlphaNumeric + 1) + val.Substring(lastAlphaNumeric + 1);
        }
        internal static string succ(string val, int length)
        {
            char lastChar = val[length - 1];
            switch (lastChar) {
                case '9':
                    return ((length > 1) ? succ(val, length - 1) : "1") + '0';
                case 'z':
                    return ((length > 1) ? succ(val, length - 1) : "a") + 'a';
                case 'Z':
                    return ((length > 1) ? succ(val, length - 1) : "A") + 'A';
                default:
                    return val.Substring(0, length - 1) + (char)(lastChar + 1);
        }
        #endregion
    }

        #endregion

        #region Overloaded creation methods, for specific range types
    public static IEnumerable<DateTime> Create(DateTime start, DateTime finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<string> Create(string start, string finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<int> Create(int start, int finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<long> Create(long start, long finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<short> Create(short start, short finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<sbyte> Create(sbyte start, sbyte finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<ulong> Create(ulong start, ulong finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<uint> Create(uint start, uint finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<ushort> Create(ushort start, ushort finish)
    { return Create(start, finish, succ, comp); }

    public static IEnumerable<byte> Create(byte start, byte finish)
    { return Create(start, finish, succ, comp); }
    #endregion

    }

}
