using System;
using System.Collections.Generic;
using System.Text;
using LINQQueries;

namespace MySpecification
{
    public interface Specification<T>
    {
        bool IsSatisfiedBy(T customer);
        Specification<T> And(Specification<T> other); 
        Specification<T> Or(Specification<T> other); 
        Specification<T> Not();
    } 

    public abstract class AbstractSpecification<T>: Specification<T>
    {
       public Specification<T> And(Specification<T> other) 
       {
          return new AndSpecification<T>(this, other);
       }
       public Specification<T> Or(Specification<T> other)
       {
          return new OrSpecification<T>(this, other);
       }
       public Specification<T> Not()
       {
          return new NotSpecification<T>(this);
       }

        abstract public bool IsSatisfiedBy(T customer);
   }

    public class AndSpecification<T> : AbstractSpecification<T>
    {
       private Specification<T> rhs;
       private Specification<T> lhs;
       public AndSpecification(Specification<T> rhs, Specification<T> lhs)
       {
           this.rhs = rhs;
           this.lhs = lhs;
       }

       public override bool IsSatisfiedBy(T candidate)
       {
          return rhs.IsSatisfiedBy(candidate) &&
             lhs.IsSatisfiedBy(candidate);
       }
    }

   public class OrSpecification<T> : AbstractSpecification<T>
    {
       private Specification<T> rhs;
       private Specification<T> lhs;
       public OrSpecification(Specification<T> rhs, Specification<T> lhs)
       {
           this.rhs = rhs;
           this.lhs = lhs;
       }

       public override bool IsSatisfiedBy(T candidate)
       {
          return rhs.IsSatisfiedBy(candidate) ||
             lhs.IsSatisfiedBy(candidate);
       }
    }

   public class NotSpecification<T> : AbstractSpecification<T>
    {
        private Specification<T> spec;
        public NotSpecification(Specification<T> spec)
        {
           this.spec = spec;
        }

        public override bool IsSatisfiedBy(T candidate)
        {
              return !spec.IsSatisfiedBy(candidate);
        }
    }


    public class CustomerCitySpecification : AbstractSpecification<Customer>
    {
        private string city;

        /// <summary>
        /// Initializes a new instance of the CustomerForCitySpecification class.
        /// </summary>
        /// <param name="city"></param>
        public CustomerCitySpecification(string city)
        {
            this.city = city;
        }

        public override bool IsSatisfiedBy(Customer candidate)
        {
            return candidate.City == city;
        }
    }

    public class CustomerCountrySpecification : AbstractSpecification<Customer>
    {
        private string country;

        /// <summary>
        /// Initializes a new instance of the CustomerForCitySpecification class.
        /// </summary>
        /// <param name="city"></param>
        public CustomerCountrySpecification(string country)
        {
            this.country = country;
        }

        public override bool IsSatisfiedBy(Customer candidate)
        {
            return candidate.Country == country;
        }
    }
}
