﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Query;
using System.Xml.XLinq;
using System.Data.DLinq;
using LINQQueries;
using MySpecification;
using System.Expressions;

namespace Extensions
{
   public static class Spec
   {
        public static List<Customer> SelectSatisfying(this List<Customer> existingCustomers, ISpecification specification)
        {
            List<Customer> customers = new List<Customer>();
            foreach (Customer customer in existingCustomers)
            {
                if (specification.IsSatisfiedBy(customer))
                    customers.Add(customer);
            }

            return customers;
        }

        public static List<Customer> SelectSatisfying(this List<Customer> existingCustomers, Func<Customer, string, bool> specification, string city)
        {
            List<Customer> customers = new List<Customer>();
            foreach (Customer customer in existingCustomers)
            {
                if (specification(customer, city))
                    customers.Add(customer);
            }

            return customers;
        }

        public static List<Customer> SelectSatisfying(this List<Customer> existingCustomers, Specification<Customer> specification)
        {
            List<Customer> customers = new List<Customer>();
            foreach (Customer customer in existingCustomers)
            {
                if (specification.IsSatisfiedBy(customer))
                    customers.Add(customer);
            }

            return customers;
        }

       public static Expression<Predicate<T>> And<T>(this Expression<Predicate<T>> rhs, Expression<Predicate<T>> lhs)
       {
           //Parameters to both expressions need to be the same and of type T
           ParameterExpression parameter = Expression.Parameter(typeof(T), "p");
                                 
           var invokedExpression = Expression.Invoke (lhs, rhs.Parameters.Cast<Expression> ());
           return Expression.Lambda<Predicate<T>> (Expression.And (rhs.Body, invokedExpression), rhs.Parameters);
       }

        public static List<Customer> SelectSatisfying(this List<Customer> existingCustomers, IQuickSpecification<Customer> specification)
        {
            List<Customer> customers = new List<Customer>();
            foreach (Customer customer in existingCustomers)
            {
                if (specification.IsSatisfiedBy(customer))
                    customers.Add(customer);
            }

            return customers;
        }

   }
}

namespace MySpecification
{
    using Extensions;
    
    class Program
    {
        static void Main(string[] args)
        {
            DataSource dataSource = new DataSource();

            #region Hard Coded Specification
            //List<Customer> londonCustomers = dataSource.GetCustomerList().SelectSatisfying(new CustomerInLondonSpecification());
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

            #region Hard Coded Specification via Lambda Expression
            //Predicate<Customer> isCustomerInLondon = c => c.City == "London";
            //List<Customer> londonCustomers = dataSource.GetCustomerList().FindAll(isCustomerInLondon);
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

            #region LINQ is a specification
            //var query =
            //    from c in dataSource.GetCustomerList()
            //    where c.City == "London"
            //    select c;

            //ObjectDumper.Write(query, 1);
            #endregion

            #region Parameterized specification
            //ISpecification isInLondon = new CustomerForCitySpecification("London");
            //List<Customer> londonCustomers = dataSource.GetCustomerList().SelectSatisfying(isInLondon);
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

            #region Parameterized specification with Lambda expression
            //Func<Customer, string, bool> isCustomerIn = ( c, cty) => c.City == cty;
            //List<Customer> londonCustomers = dataSource.GetCustomerList().SelectSatisfying(isCustomerIn, "London");
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

            #region Parameterized composite specification
            //Specification<Customer> isInLondon = new CustomerCitySpecification("London").And(new CustomerCountrySpecification("UK"));
            //List<Customer> londonCustomers = dataSource.GetCustomerList().SelectSatisfying(isInLondon);
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

            #region Hardcoded composition
            //Predicate<Customer> isCustomerInLondon = c => c.City == "London" && c.Country ==  "UK";
            //List<Customer> londonCustomers = dataSource.GetCustomerList().FindAll(isCustomerInLondon);
            //ObjectDumper.Write(londonCustomers, 1);

            #endregion

            #region Dynamic composition of predicates
            //Expression<Predicate<Customer>> isCustomerInLondon = c => c.City == "London";
            //Expression<Predicate<Customer>> isInUK = c => c.Country == "UK";
            //Expression<Predicate<Customer>> isLondonUKCustomer = isCustomerInLondon.And(isInUK);
            //List<Customer> londonCustomers = dataSource.GetCustomerList().FindAll(isLondonUKCustomer.Compile());
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

            #region Composite Lambda specifications build using expression trees
            //IQuickSpecification<Customer> isCustomerInLondon = new QuickSpecification<Customer>(c => c.City == "London");
            //IQuickSpecification<Customer> isInUK = new QuickSpecification<Customer>(c => c.Country == "UK");
            //IQuickSpecification<Customer> isLondonUKCustomer = isCustomerInLondon.And(isInUK);
            //List<Customer> londonCustomers = dataSource.GetCustomerList().SelectSatisfying(isLondonUKCustomer);
            //ObjectDumper.Write(londonCustomers, 1);
            #endregion

        }
    }
}
